@@ -170,14 +170,6 @@ bool ExpressionAnalyzer::flexibleEqual(Expression* left, Expression* right, Expr
         }
         break;
       }
-      case Expression::Id::CallImportId: {
-        CHECK(CallImport, target);
-        CHECK(CallImport, operands.size());
-        for (Index i = 0; i < left->cast<CallImport>()->operands.size(); i++) {
-          PUSH(CallImport, operands[i]);
-        }
-        break;
-      }
       case Expression::Id::CallIndirectId: {
         PUSH(CallIndirect, target);
         CHECK(CallIndirect, fullType);
@@ -257,7 +249,7 @@ bool ExpressionAnalyzer::flexibleEqual(Expression* left, Expression* right, Expr
         break;
       }
       case Expression::Id::ConstId: {
-        if (!left->cast<Const>()->value.bitwiseEqual(right->cast<Const>()->value)) {
+        if (left->cast<Const>()->value != right->cast<Const>()->value) {
           return false;
         }
         break;
@@ -423,14 +415,6 @@ uint32_t ExpressionAnalyzer::hash(Expression* curr) {
         }
         break;
       }
-      case Expression::Id::CallImportId: {
-        HASH_NAME(CallImport, target);
-        HASH(CallImport, operands.size());
-        for (Index i = 0; i < curr->cast<CallImport>()->operands.size(); i++) {
-          PUSH(CallImport, operands[i]);
-        }
-        break;
-      }
       case Expression::Id::CallIndirectId: {
         PUSH(CallIndirect, target);
         HASH_NAME(CallIndirect, fullType);
@@ -510,8 +494,14 @@ uint32_t ExpressionAnalyzer::hash(Expression* curr) {
         break;
       }
       case Expression::Id::ConstId: {
-        HASH(Const, value.type);
-        HASH64(Const, value.getBits());
+        auto* c = curr->cast<Const>();
+        hash(c->type);
+        auto bits = c->value.getBits();
+        if (getTypeSize(c->type) == 4) {
+          hash(uint32_t(bits));
+        } else {
+          hash64(bits);
+        }
         break;
       }
       case Expression::Id::UnaryId: {
