@@ -32,6 +32,7 @@
 #include "wasm-builder.h"
 #include "parsing.h"
 #include "wasm-validator.h"
+#include "ir/import-utils.h"
 
 namespace wasm {
 
@@ -44,7 +45,8 @@ enum {
 // accept
 enum WebLimitations {
   MaxDataSegments = 100 * 1000,
-  MaxFunctionBodySize = 128 * 1024
+  MaxFunctionBodySize = 128 * 1024,
+  MaxFunctionLocals = 50 * 1000
 };
 
 template<typename T, typename MiniT>
@@ -340,6 +342,8 @@ namespace UserSections {
 extern const char* Name;
 extern const char* SourceMapUrl;
 
+extern const char* Dylink;
+
 enum Subsection {
   NameFunction = 1,
   NameLocal = 2,
@@ -652,21 +656,14 @@ inline S32LEB binaryType(Type type) {
   return S32LEB(ret);
 }
 
-class WasmBinaryWriter : public Visitor<WasmBinaryWriter, void> {
-  Module* wasm;
-  BufferWithRandomAccess& o;
-  Function* currFunction = nullptr;
-  bool debug;
-  bool debugInfo = true;
-  std::ostream* sourceMap = nullptr;
-  std::string sourceMapUrl;
-  std::string symbolMap;
+// Writes out wasm to the binary format
 
-  MixedArena allocator;
-
-  void prepare();
+class WasmBinaryWriter {
 public:
-  WasmBinaryWriter(Module* input, BufferWithRandomAccess& o, bool debug = false) : wasm(input), o(o), debug(debug) {
+  WasmBinaryWriter(Module* input,
+                   BufferWithRandomAccess& o,
+                   bool debug = false) :
+      wasm(input), o(o), debug(debug) {
     prepare();
   }
 
@@ -703,10 +700,6 @@ public:
   int32_t getFunctionTypeIndex(Name type);
   void writeImports();
 
-  std::map<Index, size_t> mappedLocals; // local index => index in compact form of [all int32s][all int64s]etc
-  std::map<Type, size_t> numLocalsByType; // type => number of locals of that type in the compact form
-
-  void mapLocals(Function* function);
   void writeFunctionSignatures();
   void writeExpression(Expression* curr);
   void writeFunctions();
@@ -724,14 +717,18 @@ public:
   void writeNames();
   void writeSourceMapUrl();
   void writeSymbolMap();
-  void writeUserSections();
+  void writeEarlyUserSections();
+  void writeLateUserSections();
+  void writeUserSection(const UserSection& section);
 
   void writeSourceMapProlog();
   void writeSourceMapEpilog();
-  void writeDebugLocation(size_t offset, const Function::DebugLocation& loc);
+  void writeDebugLocation(const Function::DebugLocation& loc);
+  void writeDebugLocation(Expression* curr, Function* func);
 
   // helpers
   void writeInlineString(const char* name);
+  void writeEscapedName(const char* name);
   void writeInlineBuffer(const char* data, size_t size);
 
   struct Buffer {
@@ -747,57 +744,29 @@ public:
   void emitString(const char *str);
   void finishUp();
 
-  // AST writing via visitors
-  int depth = 0; // only for debugging
+  Module* getModule() { return wasm; }
 
-  void recurse(Expression*& curr);
-  std::vector<Name> breakStack;
+private:
+  Module* wasm;
+  BufferWithRandomAccess& o;
+  bool debug;
+
+  bool debugInfo = true;
+  std::ostream* sourceMap = nullptr;
+  std::string sourceMapUrl;
+  std::string symbolMap;
+
+  MixedArena allocator;
+
+  // storage of source map locations until the section is placed at its final location
+  // (shrinking LEBs may cause changes there)
+  std::vector<std::pair<size_t, const Function::DebugLocation*>> sourceMapLocations;
+  size_t sourceMapLocationsSizeAtSectionStart;
   Function::DebugLocation lastDebugLocation;
-  size_t lastBytecodeOffset;
 
-  void visit(Expression* curr) {
-    if (sourceMap && currFunction) {
-      // Dump the sourceMap debug info
-      auto& debugLocations = currFunction->debugLocations;
-      auto iter = debugLocations.find(curr);
-      if (iter != debugLocations.end() && iter->second != lastDebugLocation) {
-        writeDebugLocation(o.size(), iter->second);
-      }
-    }
-    Visitor<WasmBinaryWriter>::visit(curr);
-  }
+  std::unique_ptr<ImportInfo> importInfo;
 
-  void visitBlock(Block *curr);
-  // emits a node, but if it is a block with no name, emit a list of its contents
-  void recursePossibleBlockContents(Expression* curr);
-  void visitIf(If *curr);
-  void visitLoop(Loop *curr);
-  int32_t getBreakIndex(Name name);
-  void visitBreak(Break *curr);
-  void visitSwitch(Switch *curr);
-  void visitCall(Call *curr);
-  void visitCallImport(CallImport *curr);
-  void visitCallIndirect(CallIndirect *curr);
-  void visitGetLocal(GetLocal *curr);
-  void visitSetLocal(SetLocal *curr);
-  void visitGetGlobal(GetGlobal *curr);
-  void visitSetGlobal(SetGlobal *curr);
-  void emitMemoryAccess(size_t alignment, size_t bytes, uint32_t offset);
-  void visitLoad(Load *curr);
-  void visitStore(Store *curr);
-  void visitAtomicRMW(AtomicRMW *curr);
-  void visitAtomicCmpxchg(AtomicCmpxchg *curr);
-  void visitAtomicWait(AtomicWait *curr);
-  void visitAtomicWake(AtomicWake *curr);
-  void visitConst(Const *curr);
-  void visitUnary(Unary *curr);
-  void visitBinary(Binary *curr);
-  void visitSelect(Select *curr);
-  void visitReturn(Return *curr);
-  void visitHost(Host *curr);
-  void visitNop(Nop *curr);
-  void visitUnreachable(Unreachable *curr);
-  void visitDrop(Drop *curr);
+  void prepare();
 };
 
 class WasmBinaryBuilder {
@@ -810,7 +779,7 @@ class WasmBinaryBuilder {
 
   size_t pos = 0;
   Index startIndex = -1;
-  bool useDebugLocation;
+  std::set<Function::DebugLocation> debugLocation;
 
   std::set<BinaryConsts::Section> seenSections;
 
@@ -822,7 +791,7 @@ public:
       debug(debug),
       sourceMap(nullptr),
       nextDebugLocation(0, { 0, 0, 0 }),
-      useDebugLocation(false) {}
+      debugLocation() {}
 
   void read();
   void readUserSection(size_t payloadLen);
@@ -867,9 +836,8 @@ public:
 
   // We read functions before we know their names, so we need to backpatch the names later
   std::vector<Function*> functions; // we store functions here before wasm.addFunction after we know their names
-  std::vector<Import*> functionImports; // we store function imports here before wasm.addFunctionImport after we know their names
-  std::map<Index, std::vector<Call*>> functionCalls; // at index i we have all calls to the defined function i
-  std::map<Index, std::vector<CallImport*>> functionImportCalls; // at index i we have all callImports to the imported function i
+  std::vector<Function*> functionImports; // we store function imports here before wasm.addFunctionImport after we know their names
+  std::map<Index, std::vector<Call*>> functionCalls; // at index i we have all calls to the function i
   Function* currFunction = nullptr;
   Index endOfFunction = -1; // before we see a function (like global init expressions), there is no end of function to check
 
@@ -938,7 +906,6 @@ public:
   void setDebugLocations(std::istream* sourceMap_) {
       sourceMap = sourceMap_;
   }
-  Function::DebugLocation debugLocation;
   std::unordered_map<std::string, Index> debugInfoFileIndices;
   void readNextDebugLocation();
   void readSourceMapHeader();
@@ -948,34 +915,23 @@ public:
 
   BinaryConsts::ASTNodes readExpression(Expression*& curr);
   void pushBlockElements(Block* curr, size_t start, size_t end);
-  void visitBlock(Block *curr);
+  void visitBlock(Block* curr);
 
   // Gets a block of expressions. If it's just one, return that singleton.
   Expression* getBlockOrSingleton(Type type);
 
-  void visitIf(If *curr);
-  void visitLoop(Loop *curr);
+  void visitIf(If* curr);
+  void visitLoop(Loop* curr);
   BreakTarget getBreakTarget(int32_t offset);
   void visitBreak(Break *curr, uint8_t code);
-  void visitSwitch(Switch *curr);
+  void visitSwitch(Switch* curr);
 
-  template<typename T>
-  void fillCall(T* call, FunctionType* type) {
-    assert(type);
-    auto num = type->params.size();
-    call->operands.resize(num);
-    for (size_t i = 0; i < num; i++) {
-      call->operands[num - i - 1] = popNonVoidExpression();
-    }
-    call->type = type->result;
-  }
-
-  Expression* visitCall();
-  void visitCallIndirect(CallIndirect *curr);
-  void visitGetLocal(GetLocal *curr);
+  void visitCall(Call* curr);
+  void visitCallIndirect(CallIndirect* curr);
+  void visitGetLocal(GetLocal* curr);
   void visitSetLocal(SetLocal *curr, uint8_t code);
-  void visitGetGlobal(GetGlobal *curr);
-  void visitSetGlobal(SetGlobal *curr);
+  void visitGetGlobal(GetGlobal* curr);
+  void visitSetGlobal(SetGlobal* curr);
   void readMemoryAccess(Address& alignment, Address& offset);
   bool maybeVisitLoad(Expression*& out, uint8_t code, bool isAtomic);
   bool maybeVisitStore(Expression*& out, uint8_t code, bool isAtomic);
@@ -986,12 +942,12 @@ public:
   bool maybeVisitConst(Expression*& out, uint8_t code);
   bool maybeVisitUnary(Expression*& out, uint8_t code);
   bool maybeVisitBinary(Expression*& out, uint8_t code);
-  void visitSelect(Select *curr);
-  void visitReturn(Return *curr);
+  void visitSelect(Select* curr);
+  void visitReturn(Return* curr);
   bool maybeVisitHost(Expression*& out, uint8_t code);
-  void visitNop(Nop *curr);
-  void visitUnreachable(Unreachable *curr);
-  void visitDrop(Drop *curr);
+  void visitNop(Nop* curr);
+  void visitUnreachable(Unreachable* curr);
+  void visitDrop(Drop* curr);
 
   void throwError(std::string text);
 };
