@@ -22,9 +22,10 @@
 #include "asmjs/shared-constants.h"
 #include "shared-constants.h"
 #include "wasm-builder.h"
-#include "wasm-linker.h"
 #include "wasm-traversal.h"
 #include "wasm.h"
+#include "ir/function-type-utils.h"
+#include "ir/module-utils.h"
 
 namespace wasm {
 
@@ -159,6 +160,18 @@ static bool hasI64ResultOrParam(FunctionType* ft) {
   return false;
 }
 
+inline void exportFunction(Module& wasm, Name name, bool must_export) {
+  if (!wasm.getFunctionOrNull(name)) {
+    assert(!must_export);
+    return;
+  }
+  if (wasm.getExportOrNull(name)) return; // Already exported
+  auto exp = new Export;
+  exp->name = exp->value = name;
+  exp->kind = ExternalKind::Function;
+  wasm.addExport(exp);
+}
+
 void EmscriptenGlueGenerator::generateDynCallThunks() {
   std::unordered_set<std::string> sigs;
   Builder builder(wasm);
@@ -170,12 +183,7 @@ void EmscriptenGlueGenerator::generateDynCallThunks() {
     if (indirectFunc == dummyFunction) {
       continue;
     }
-    std::string sig;
-    if (auto import = wasm.getImportOrNull(indirectFunc)) {
-      sig = getSig(wasm.getFunctionType(import->functionType));
-    } else {
-      sig = getSig(wasm.getFunction(indirectFunc));
-    }
+    std::string sig = getSig(wasm.getFunction(indirectFunc));
     auto* funcType = ensureFunctionType(sig, &wasm);
     if (hasI64ResultOrParam(funcType)) continue; // Can't export i64s on the web.
     if (!sigs.insert(sig).second) continue; // Sig is already in the set
@@ -207,16 +215,16 @@ struct JSCallWalker : public PostWalker<JSCallWalker> {
     }
     const auto& tableSegmentData = wasm.table.segments[0].data;
 
+    jsCallStartIndex =
+        wasm.table.segments[0].offset->cast<Const>()->value.getInteger();
     // Check if jsCalls have already been created
     for (Index i = 0; i < tableSegmentData.size(); ++i) {
       if (tableSegmentData[i].startsWith("jsCall_")) {
-        jsCallStartIndex = i;
+        jsCallStartIndex += i;
         return;
       }
     }
-    jsCallStartIndex =
-        wasm.table.segments[0].offset->cast<Const>()->value.getInteger() +
-        tableSegmentData.size();
+    jsCallStartIndex += tableSegmentData.size();
   }
 
   // Gather all function signatures used in call_indirect, because any of them
@@ -258,12 +266,12 @@ void EmscriptenGlueGenerator::generateJSCallThunks(
     // function would have signature 'vii'.)
     std::string importSig = std::string(1, sig[0]) + 'i' + sig.substr(1);
     FunctionType *importType = ensureFunctionType(importSig, &wasm);
-    auto import = new Import;
+    auto import = new Function;
     import->name = import->base = "jsCall_" + sig;
     import->module = ENV;
-    import->functionType = importType->name;
-    import->kind = ExternalKind::Function;
-    wasm.addImport(import);
+    import->type = importType->name;
+    FunctionTypeUtils::fillFunction(import, importType);
+    wasm.addFunction(import);
     FunctionType *funcType = ensureFunctionType(sig, &wasm);
 
     // Create jsCall_sig_index thunks (e.g. jsCall_vi_0, jsCall_vi_1, ...)
@@ -286,7 +294,7 @@ void EmscriptenGlueGenerator::generateJSCallThunks(
         args.push_back(builder.makeGetLocal(i, funcType->params[i]));
       }
       Expression* call =
-          builder.makeCallImport(import->name, args, funcType->result);
+          builder.makeCall(import->name, args, funcType->result);
       f->body = call;
       wasm.addFunction(f);
       tableSegmentData.push_back(f->name);
@@ -367,18 +375,23 @@ struct AsmConstWalker : public PostWalker<AsmConstWalker> {
     : wasm(_wasm),
       segmentOffsets(getSegmentOffsets(wasm)) { }
 
-  void visitCallImport(CallImport* curr);
+  void visitCall(Call* curr);
+
+  void process();
 
 private:
   Literal idLiteralForCode(std::string code);
   std::string asmConstSig(std::string baseSig);
   Name nameForImportWithSig(std::string sig);
-  void addImport(Name importName, std::string baseSig);
+  void queueImport(Name importName, std::string baseSig);
+  void addImports();
+
+  std::vector<std::unique_ptr<Function>> queuedImports;
 };
 
-void AsmConstWalker::visitCallImport(CallImport* curr) {
-  Import* import = wasm.getImport(curr->target);
-  if (import->base.hasSubstring(EMSCRIPTEN_ASM_CONST)) {
+void AsmConstWalker::visitCall(Call* curr) {
+  auto* import = wasm.getFunction(curr->target);
+  if (import->imported() && import->base.hasSubstring(EMSCRIPTEN_ASM_CONST)) {
     auto arg = curr->operands[0]->cast<Const>();
     auto code = codeForConstAddr(wasm, segmentOffsets, arg);
     arg->value = idLiteralForCode(code);
@@ -390,11 +403,19 @@ void AsmConstWalker::visitCallImport(CallImport* curr) {
 
     if (allSigs.count(sig) == 0) {
       allSigs.insert(sig);
-      addImport(importName, baseSig);
+      queueImport(importName, baseSig);
     }
   }
 }
 
+void AsmConstWalker::process() {
+  // Find and queue necessary imports
+  walkModule(&wasm);
+  // Add them after the walk, to avoid iterator invalidation on
+  // the list of functions.
+  addImports();
+}
+
 Literal AsmConstWalker::idLiteralForCode(std::string code) {
   int32_t id;
   if (ids.count(code) == 0) {
@@ -422,32 +443,37 @@ Name AsmConstWalker::nameForImportWithSig(std::string sig) {
   return Name(fixedTarget.c_str());
 }
 
-void AsmConstWalker::addImport(Name importName, std::string baseSig) {
-  auto import = new Import;
+void AsmConstWalker::queueImport(Name importName, std::string baseSig) {
+  auto import = new Function;
   import->name = import->base = importName;
   import->module = ENV;
-  import->functionType = ensureFunctionType(baseSig, &wasm)->name;
-  import->kind = ExternalKind::Function;
-  wasm.addImport(import);
+  import->type = ensureFunctionType(baseSig, &wasm)->name;
+  queuedImports.push_back(std::unique_ptr<Function>(import));
+}
+
+void AsmConstWalker::addImports() {
+  for (auto& import : queuedImports) {
+    wasm.addFunction(import.release());
+  }
 }
 
 AsmConstWalker fixEmAsmConstsAndReturnWalker(Module& wasm) {
   // Collect imports to remove
   // This would find our generated functions if we ran it later
   std::vector<Name> toRemove;
-  for (auto& import : wasm.imports) {
-    if (import->base.hasSubstring(EMSCRIPTEN_ASM_CONST)) {
+  for (auto& import : wasm.functions) {
+    if (import->imported() && import->base.hasSubstring(EMSCRIPTEN_ASM_CONST)) {
       toRemove.push_back(import->name);
     }
   }
 
   // Walk the module, generate _sig versions of EM_ASM functions
   AsmConstWalker walker(wasm);
-  walker.walkModule(&wasm);
+  walker.process();
 
   // Remove the base functions that we didn't generate
   for (auto importName : toRemove) {
-    wasm.removeImport(importName);
+    wasm.removeFunction(importName);
   }
   return walker;
 }
@@ -463,6 +489,9 @@ struct EmJsWalker : public PostWalker<EmJsWalker> {
       segmentOffsets(getSegmentOffsets(wasm)) { }
 
   void visitFunction(Function* curr) {
+    if (curr->imported()) {
+      return;
+    }
     if (!curr->name.startsWith(EM_JS_PREFIX.str)) {
       return;
     }
@@ -508,6 +537,132 @@ void EmscriptenGlueGenerator::fixEmAsmConsts() {
   fixEmJsFuncsAndReturnWalker(wasm);
 }
 
+
+// Fixes function name hacks caused by LLVM exception & setjmp/longjmp
+// handling pass for wasm.
+// This does two things:
+// 1. Change emscripten_longjmp_jmpbuf to emscripten_longjmp.
+//    In setjmp/longjmp handling pass in wasm backend, what we want to do is
+//    to change all function calls to longjmp to calls to emscripten_longjmp.
+//    Because we replace all calls to longjmp to emscripten_longjmp, the
+//    signature of that function should be the same as longjmp:
+//    emscripten_longjmp(jmp_buf, int)
+//    But after calling a function that might longjmp, while we test whether
+//    a longjmp occurred, we have to load an int address value and call
+//    emscripten_longjmp again with that address as the first argument. (Refer
+//    to lib/Target/WebAssembly/WebAssemblyEmscriptenEHSjLj.cpp in LLVM for
+//    details.)
+//    In this case we need the signature of emscripten_longjmp to be (int,
+//    int). So we need two different kinds of emscripten_longjmp signatures in
+//    LLVM IR. Both signatures will be lowered to (int, int) eventually, but
+//    in LLVM IR, types are not lowered yet.
+//    So we declare two functions in LLVM:
+//    emscripten_longjmp_jmpbuf(jmp_buf, int)
+//    emscripten_longjmp(int, int)
+//    And we change the name of emscripten_longjmp_jmpbuf to
+//    emscripten_longjmp here.
+// 2. Converts invoke wrapper names.
+//    Refer to the comments in fixEmExceptionInvoke below.
+struct FixInvokeFunctionNamesWalker : public PostWalker<FixInvokeFunctionNamesWalker> {
+  Module& wasm;
+  std::map<Name, Name> importRenames;
+  std::vector<Name> toRemove;
+  std::set<Name> newImports;
+
+  FixInvokeFunctionNamesWalker(Module& _wasm) : wasm(_wasm) {}
+
+  // Converts invoke wrapper names generated by LLVM backend to real invoke
+  // wrapper names that are expected by JavaScript glue code.
+  // This is required to support wasm exception handling (asm.js style).
+  //
+  // LLVM backend lowers
+  //   invoke @func(arg1, arg2) to label %invoke.cont unwind label %lpad
+  // into
+  // ... (some code)
+  //   call @invoke_SIG(func, arg1, arg2)
+  // ... (some code)
+  // SIG is a mangled string generated based on the LLVM IR-level function
+  // signature. In LLVM IR, types are not lowered yet, so this mangling scheme
+  // simply takes LLVM's string representtion of parameter types and concatenate
+  // them with '_'. For example, the name of an invoke wrapper for function
+  // void foo(struct mystruct*, int) will be
+  // "__invoke_void_%struct.mystruct*_int".
+  // This function converts the names of invoke wrappers based on their lowered
+  // argument types and a return type. In the example above, the resulting new
+  // wrapper name becomes "invoke_vii".
+  static Name fixEmExceptionInvoke(const Name& name, const std::string& sig) {
+    std::string nameStr = name.c_str();
+    if (nameStr.front() == '"' && nameStr.back() == '"') {
+      nameStr = nameStr.substr(1, nameStr.size() - 2);
+    }
+    if (nameStr.find("__invoke_") != 0) {
+      return name;
+    }
+    std::string sigWoOrigFunc = sig.front() + sig.substr(2, sig.size() - 2);
+    return Name("invoke_" + sigWoOrigFunc);
+  }
+
+  static Name fixEmEHSjLjNames(const Name &name, const std::string& sig) {
+    if (name == "emscripten_longjmp_jmpbuf")
+      return "emscripten_longjmp";
+    return fixEmExceptionInvoke(name, sig);
+  }
+
+  void visitFunction(Function* curr) {
+    if (!curr->imported()) {
+      return;
+    }
+
+    FunctionType* func = wasm.getFunctionType(curr->type);
+    Name newname = fixEmEHSjLjNames(curr->base, getSig(func));
+    if (newname == curr->base) {
+      return;
+    }
+
+    assert(importRenames.count(curr->name) == 0);
+    importRenames[curr->name] = newname;
+    // Either rename or remove the existing import
+    if (wasm.getFunctionOrNull(newname) || !newImports.insert(newname).second) {
+      toRemove.push_back(curr->name);
+    } else {
+      curr->base = newname;
+      curr->name = newname;
+    }
+  }
+
+  void visitTable(Table* curr) {
+    for (auto& segment : curr->segments) {
+      for (size_t i = 0; i < segment.data.size(); i++) {
+        auto it = importRenames.find(segment.data[i]);
+        if (it != importRenames.end()) {
+          segment.data[i] = it->second;
+        }
+      }
+    }
+  }
+
+  void visitCall(Call* curr) {
+    if (wasm.getFunction(curr->target)->imported()) {
+      auto it = importRenames.find(curr->target);
+      if (it != importRenames.end()) {
+        curr->target = it->second;
+      }
+    }
+  }
+
+  void visitModule(Module* curr) {
+    for (auto importName : toRemove) {
+      wasm.removeFunction(importName);
+    }
+    wasm.updateMaps();
+  }
+};
+
+void EmscriptenGlueGenerator::fixInvokeFunctionNames() {
+  FixInvokeFunctionNamesWalker walker(wasm);
+  walker.walkModule(&wasm);
+}
+
 template<class C>
 void printSet(std::ostream& o, C& c) {
   o << "[";
@@ -587,32 +742,42 @@ std::string EmscriptenGlueGenerator::generateEmscriptenMetadata(
     meta << "]";
   }
 
+  // Avoid adding duplicate imports to `declares' or `invokeFuncs`.  Even
+  // though we might import the same function multiple times (i.e. with
+  // different sigs) we only need to list is in the metadata once.
+  std::set<std::string> declares;
+  std::set<std::string> invokeFuncs;
+
+  // We use the `base` rather than the `name` of the imports here and below
+  // becasue this is the externally visible name that the embedder (JS) will
+  // see.
   meta << ", \"declares\": [";
   commaFirst = true;
-  for (const auto& import : wasm.imports) {
-    if (import->kind == ExternalKind::Function &&
-        (emJsWalker.codeByName.count(import->name.str) == 0) &&
-        !import->name.startsWith(EMSCRIPTEN_ASM_CONST.str) &&
-        !import->name.startsWith("invoke_") &&
-        !import->name.startsWith("jsCall_")) {
-      meta << maybeComma() << '"' << import->name.str << '"';
+  ModuleUtils::iterImportedFunctions(wasm, [&](Function* import) {
+    if (emJsWalker.codeByName.count(import->base.str) == 0 &&
+        !import->base.startsWith(EMSCRIPTEN_ASM_CONST.str) &&
+        !import->base.startsWith("invoke_") &&
+        !import->base.startsWith("jsCall_")) {
+      if (declares.insert(import->base.str).second) {
+        meta << maybeComma() << '"' << import->base.str << '"';
+      }
     }
-  }
+  });
   meta << "]";
 
   meta << ", \"externs\": [";
   commaFirst = true;
-  for (const auto& import : wasm.imports) {
-    if (import->kind == ExternalKind::Global) {
-      meta << maybeComma() << "\"_" << import->name.str << '"';
-    }
-  }
+  ModuleUtils::iterImportedGlobals(wasm, [&](Global* import) {
+    meta << maybeComma() << "\"_" << import->base.str << '"';
+  });
   meta << "]";
 
   meta << ", \"implementedFunctions\": [";
   commaFirst = true;
-  for (const auto& func : wasm.functions) {
-    meta << maybeComma() << "\"_" << func->name.str << '"';
+  for (const auto& ex : wasm.exports) {
+    if (ex->kind == ExternalKind::Function) {
+      meta << maybeComma() << "\"_" << ex->name.str << '"';
+    }
   }
   meta << "]";
 
@@ -625,11 +790,13 @@ std::string EmscriptenGlueGenerator::generateEmscriptenMetadata(
 
   meta << ", \"invokeFuncs\": [";
   commaFirst = true;
-  for (const auto& import : wasm.imports) {
-    if (import->name.startsWith("invoke_")) {
-      meta << maybeComma() << '"' << import->name.str << '"';
+  ModuleUtils::iterImportedFunctions(wasm, [&](Function* import) {
+    if (import->base.startsWith("invoke_")) {
+      if (invokeFuncs.insert(import->base.str).second) {
+        meta << maybeComma() << '"' << import->base.str << '"';
+      }
     }
-  }
+  });
   meta << "]";
 
   meta << " }\n";
@@ -637,28 +804,4 @@ std::string EmscriptenGlueGenerator::generateEmscriptenMetadata(
   return meta.str();
 }
 
-std::string emscriptenGlue(
-    Module& wasm,
-    bool allowMemoryGrowth,
-    Address stackPointer,
-    Address staticBump,
-    std::vector<Name> const& initializerFunctions,
-    unsigned numReservedFunctionPointers) {
-  EmscriptenGlueGenerator generator(wasm, stackPointer);
-  generator.generateRuntimeFunctions();
-
-  if (allowMemoryGrowth) {
-    generator.generateMemoryGrowthFunction();
-  }
-
-  generator.generateDynCallThunks();
-
-  if (numReservedFunctionPointers) {
-    generator.generateJSCallThunks(numReservedFunctionPointers);
-  }
-
-  return generator.generateEmscriptenMetadata(staticBump, initializerFunctions,
-                                              numReservedFunctionPointers);
-}
-
 } // namespace wasm
