@@ -24,6 +24,7 @@
 #include "wasm-validator.h"
 #include "ir/utils.h"
 #include "ir/branch-utils.h"
+#include "ir/module-utils.h"
 #include "support/colors.h"
 
 
@@ -171,25 +172,34 @@ struct FunctionValidator : public WalkerPass<PostWalker<FunctionValidator>> {
 
   Pass* create() override { return new FunctionValidator(&info); }
 
+  bool modifiesBinaryenIR() override { return false; }
+
   ValidationInfo& info;
 
   FunctionValidator(ValidationInfo* info) : info(*info) {}
 
   struct BreakInfo {
+    enum {
+      UnsetArity = Index(-1),
+      PoisonArity = Index(-2)
+    };
+
     Type type;
     Index arity;
-    BreakInfo() {}
+    BreakInfo() : arity(UnsetArity) {}
     BreakInfo(Type type, Index arity) : type(type), arity(arity) {}
+
+    bool hasBeenSet() {
+      // Compare to the impossible value.
+      return arity != UnsetArity;
+    }
   };
 
-  std::map<Name, Expression*> breakTargets;
-  std::map<Expression*, BreakInfo> breakInfos;
+  std::unordered_map<Name, BreakInfo> breakInfos;
 
   Type returnType = unreachable; // type used in returns
 
-  std::set<Name> labelNames; // Binaryen IR requires that label names must be unique - IR generators must ensure that
-
-  std::unordered_set<Expression*> seenExpressions; // expressions must not appear twice
+  std::unordered_set<Name> labelNames; // Binaryen IR requires that label names must be unique - IR generators must ensure that
 
   void noteLabelName(Name name);
 
@@ -198,14 +208,14 @@ public:
 
   static void visitPreBlock(FunctionValidator* self, Expression** currp) {
     auto* curr = (*currp)->cast<Block>();
-    if (curr->name.is()) self->breakTargets[curr->name] = curr;
+    if (curr->name.is()) self->breakInfos[curr->name];
   }
 
   void visitBlock(Block* curr);
 
   static void visitPreLoop(FunctionValidator* self, Expression** currp) {
     auto* curr = (*currp)->cast<Loop>();
-    if (curr->name.is()) self->breakTargets[curr->name] = curr;
+    if (curr->name.is()) self->breakInfos[curr->name];
   }
 
   void visitLoop(Loop* curr);
@@ -224,7 +234,6 @@ public:
   void visitBreak(Break* curr);
   void visitSwitch(Switch* curr);
   void visitCall(Call* curr);
-  void visitCallImport(CallImport* curr);
   void visitCallIndirect(CallIndirect* curr);
   void visitGetLocal(GetLocal* curr);
   void visitSetLocal(SetLocal* curr);
@@ -285,16 +294,19 @@ private:
 
 void FunctionValidator::noteLabelName(Name name) {
   if (!name.is()) return;
-  shouldBeTrue(labelNames.find(name) == labelNames.end(), name, "names in Binaryen IR must be unique - IR generators must ensure that");
-  labelNames.insert(name);
+  bool inserted;
+  std::tie(std::ignore, inserted) = labelNames.insert(name);
+  shouldBeTrue(inserted, name, "names in Binaryen IR must be unique - IR generators must ensure that");
 }
 
 void FunctionValidator::visitBlock(Block* curr) {
   // if we are break'ed to, then the value must be right for us
   if (curr->name.is()) {
     noteLabelName(curr->name);
-    if (breakInfos.count(curr) > 0) {
-      auto& info = breakInfos[curr];
+    auto iter = breakInfos.find(curr->name);
+    assert(iter != breakInfos.end()); // we set it ourselves
+    auto& info = iter->second;
+    if (info.hasBeenSet()) {
       if (isConcreteType(curr->type)) {
         shouldBeTrue(info.arity != 0, curr, "break arities must be > 0 if block has a value");
       } else {
@@ -307,7 +319,7 @@ void FunctionValidator::visitBlock(Block* curr) {
       if (isConcreteType(curr->type) && info.arity && info.type != unreachable) {
         shouldBeEqual(curr->type, info.type, curr, "block+breaks must have right type if breaks have arity");
       }
-      shouldBeTrue(info.arity != Index(-1), curr, "break arities must match");
+      shouldBeTrue(info.arity != BreakInfo::PoisonArity, curr, "break arities must match");
       if (curr->list.size() > 0) {
         auto last = curr->list.back()->type;
         if (isConcreteType(last) && info.type != unreachable) {
@@ -318,7 +330,7 @@ void FunctionValidator::visitBlock(Block* curr) {
         }
       }
     }
-    breakTargets.erase(curr->name);
+    breakInfos.erase(iter);
   }
   if (curr->list.size() > 1) {
     for (Index i = 0; i < curr->list.size() - 1; i++) {
@@ -347,11 +359,13 @@ void FunctionValidator::visitBlock(Block* curr) {
 void FunctionValidator::visitLoop(Loop* curr) {
   if (curr->name.is()) {
     noteLabelName(curr->name);
-    breakTargets.erase(curr->name);
-    if (breakInfos.count(curr) > 0) {
-      auto& info = breakInfos[curr];
+    auto iter = breakInfos.find(curr->name);
+    assert(iter != breakInfos.end()); // we set it ourselves
+    auto& info = iter->second;
+    if (info.hasBeenSet()) {
       shouldBeEqual(info.arity, Index(0), curr, "breaks to a loop cannot pass a value");
     }
+    breakInfos.erase(iter);
   }
   if (curr->type == none) {
     shouldBeFalse(isConcreteType(curr->body->type), curr, "bad body for a loop that has no value");
@@ -394,12 +408,12 @@ void FunctionValidator::noteBreak(Name name, Expression* value, Expression* curr
     shouldBeUnequal(valueType, none, curr, "breaks must have a valid value");
     arity = 1;
   }
-  if (!shouldBeTrue(breakTargets.count(name) > 0, curr, "all break targets must be valid")) return;
-  auto* target = breakTargets[name];
-  if (breakInfos.count(target) == 0) {
-    breakInfos[target] = BreakInfo(valueType, arity);
+  auto iter = breakInfos.find(name);
+  if (!shouldBeTrue(iter != breakInfos.end(), curr, "all break targets must be valid")) return;
+  auto& info = iter->second;
+  if (!info.hasBeenSet()) {
+    info = BreakInfo(valueType, arity);
   } else {
-    auto& info = breakInfos[target];
     if (info.type == unreachable) {
       info.type = valueType;
     } else if (valueType != unreachable) {
@@ -408,7 +422,7 @@ void FunctionValidator::noteBreak(Name name, Expression* value, Expression* curr
       }
     }
     if (arity != info.arity) {
-      info.arity = Index(-1); // a poison value
+      info.arity = BreakInfo::PoisonArity;
     }
   }
 }
@@ -430,12 +444,7 @@ void FunctionValidator::visitSwitch(Switch* curr) {
 void FunctionValidator::visitCall(Call* curr) {
   if (!info.validateGlobally) return;
   auto* target = getModule()->getFunctionOrNull(curr->target);
-  if (!shouldBeTrue(!!target, curr, "call target must exist")) {
-    if (getModule()->getImportOrNull(curr->target) && !info.quiet) {
-      getStream() << "(perhaps it should be a CallImport instead of Call?)\n";
-    }
-    return;
-  }
+  if (!shouldBeTrue(!!target, curr, "call target must exist")) return;
   if (!shouldBeTrue(curr->operands.size() == target->params.size(), curr, "call param number must match")) return;
   for (size_t i = 0; i < curr->operands.size(); i++) {
     if (!shouldBeEqualOrFirstIsUnreachable(curr->operands[i]->type, target->params[i], curr, "call param types must match") && !info.quiet) {
@@ -444,20 +453,6 @@ void FunctionValidator::visitCall(Call* curr) {
   }
 }
 
-void FunctionValidator::visitCallImport(CallImport* curr) {
-  if (!info.validateGlobally) return;
-  auto* import = getModule()->getImportOrNull(curr->target);
-  if (!shouldBeTrue(!!import, curr, "call_import target must exist")) return;
-  if (!shouldBeTrue(!!import->functionType.is(), curr, "called import must be function")) return;
-  auto* type = getModule()->getFunctionType(import->functionType);
-  if (!shouldBeTrue(curr->operands.size() == type->params.size(), curr, "call param number must match")) return;
-  for (size_t i = 0; i < curr->operands.size(); i++) {
-    if (!shouldBeEqualOrFirstIsUnreachable(curr->operands[i]->type, type->params[i], curr, "call param types must match") && !info.quiet) {
-      getStream() << "(on argument " << i << ")\n";
-    }
-  }
-}
-
 void FunctionValidator::visitCallIndirect(CallIndirect* curr) {
   if (!info.validateGlobally) return;
   auto* type = getModule()->getFunctionTypeOrNull(curr->fullType);
@@ -474,6 +469,7 @@ void FunctionValidator::visitCallIndirect(CallIndirect* curr) {
 void FunctionValidator::visitGetLocal(GetLocal* curr) {
   shouldBeTrue(curr->index < getFunction()->getNumLocals(), curr, "get_local index must be small enough");
   shouldBeTrue(isConcreteType(curr->type), curr, "get_local must have a valid type - check what you provided when you constructed the node");
+  shouldBeTrue(curr->type == getFunction()->getLocalType(curr->index), curr, "get_local must have proper type");
 }
 
 void FunctionValidator::visitSetLocal(SetLocal* curr) {
@@ -488,13 +484,13 @@ void FunctionValidator::visitSetLocal(SetLocal* curr) {
 
 void FunctionValidator::visitGetGlobal(GetGlobal* curr) {
   if (!info.validateGlobally) return;
-  shouldBeTrue(getModule()->getGlobalOrNull(curr->name) || getModule()->getImportOrNull(curr->name), curr, "get_global name must be valid");
+  shouldBeTrue(getModule()->getGlobalOrNull(curr->name), curr, "get_global name must be valid");
 }
 
 void FunctionValidator::visitSetGlobal(SetGlobal* curr) {
   if (!info.validateGlobally) return;
   auto* global = getModule()->getGlobalOrNull(curr->name);
-  if (shouldBeTrue(global != NULL, curr, "set_global name must be valid (and not an import; imports can't be modified)")) {
+  if (shouldBeTrue(global, curr, "set_global name must be valid (and not an import; imports can't be modified)")) {
     shouldBeTrue(global->mutable_, curr, "set_global global must be mutable");
     shouldBeEqualOrFirstIsUnreachable(curr->value->type, global->type, curr, "set_global value must have right type");
   }
@@ -506,7 +502,10 @@ void FunctionValidator::visitLoad(Load* curr) {
   validateMemBytes(curr->bytes, curr->type, curr);
   validateAlignment(curr->align, curr->type, curr->bytes, curr->isAtomic, curr);
   shouldBeEqualOrFirstIsUnreachable(curr->ptr->type, i32, curr, "load pointer type must be i32");
-  if (curr->isAtomic) shouldBeFalse(curr->signed_, curr, "atomic loads must be unsigned");
+  if (curr->isAtomic) {
+    shouldBeFalse(curr->signed_, curr, "atomic loads must be unsigned");
+    shouldBeIntOrUnreachable(curr->type, curr, "atomic loads must be of integers");
+  }
 }
 
 void FunctionValidator::visitStore(Store* curr) {
@@ -517,6 +516,9 @@ void FunctionValidator::visitStore(Store* curr) {
   shouldBeEqualOrFirstIsUnreachable(curr->ptr->type, i32, curr, "store pointer type must be i32");
   shouldBeUnequal(curr->value->type, none, curr, "store value type must not be none");
   shouldBeEqualOrFirstIsUnreachable(curr->value->type, curr->valueType, curr, "store value type must match");
+  if (curr->isAtomic) {
+    shouldBeIntOrUnreachable(curr->valueType, curr, "atomic stores must be of integers");
+  }
 }
 
 void FunctionValidator::visitAtomicRMW(AtomicRMW* curr) {
@@ -787,9 +789,7 @@ void FunctionValidator::visitHost(Host* curr) {
       shouldBeEqualOrFirstIsUnreachable(curr->operands[0]->type, i32, curr, "grow_memory must have i32 operand");
       break;
     }
-    case PageSize:
-    case CurrentMemory:
-    case HasFeature: break;
+    case CurrentMemory: break;
     default: WASM_UNREACHABLE();
   }
 }
@@ -809,7 +809,7 @@ void FunctionValidator::visitFunction(Function* curr) {
   if (returnType != unreachable) {
     shouldBeEqual(curr->result, returnType, curr->body, "function result must match, if function has returns");
   }
-  shouldBeTrue(breakTargets.empty(), curr->body, "all named break targets must exist");
+  shouldBeTrue(breakInfos.empty(), curr->body, "all named break targets must exist");
   returnType = unreachable;
   labelNames.clear();
   // if function has a named type, it must match up with the function's params and result
@@ -818,24 +818,6 @@ void FunctionValidator::visitFunction(Function* curr) {
     shouldBeTrue(ft->params == curr->params, curr->name, "function params must match its declared type");
     shouldBeTrue(ft->result == curr->result, curr->name, "function result must match its declared type");
   }
-  // expressions must not be seen more than once
-  struct Walker : public PostWalker<Walker, UnifiedExpressionVisitor<Walker>> {
-    std::unordered_set<Expression*>& seen;
-    std::vector<Expression*> dupes;
-
-    Walker(std::unordered_set<Expression*>& seen) : seen(seen) {}
-
-    void visitExpression(Expression* curr) {
-      bool inserted;
-      std::tie(std::ignore, inserted) = seen.insert(curr);
-      if (!inserted) dupes.push_back(curr);
-    }
-  };
-  Walker walker(seenExpressions);
-  walker.walk(curr->body);
-  for (auto* bad : walker.dupes) {
-    info.fail("expression seen more than once in the tree", bad, getFunction());
-  }
 }
 
 static bool checkOffset(Expression* curr, Address add, Address max) {
@@ -889,9 +871,12 @@ static void validateBinaryenIR(Module& wasm, ValidationInfo& info) {
   struct BinaryenIRValidator : public PostWalker<BinaryenIRValidator, UnifiedExpressionVisitor<BinaryenIRValidator>> {
     ValidationInfo& info;
 
+    std::unordered_set<Expression*> seen;
+
     BinaryenIRValidator(ValidationInfo& info) : info(info) {}
 
     void visitExpression(Expression* curr) {
+      auto scope = getFunction() ? getFunction()->name : Name("(global scope)");
       // check if a node type is 'stale', i.e., we forgot to finalize() the node.
       auto oldType = curr->type;
       ReFinalizeNode().visit(curr);
@@ -906,11 +891,19 @@ static void validateBinaryenIR(Module& wasm, ValidationInfo& info) {
         // ok for it to be either i32 or unreachable.
         if (!(isConcreteType(oldType) && newType == unreachable)) {
           std::ostringstream ss;
-          ss << "stale type found in " << (getFunction() ? getFunction()->name : Name("(global scope)")) << " on " << curr << "\n(marked as " << printType(oldType) << ", should be " << printType(newType) << ")\n";
+          ss << "stale type found in " << scope << " on " << curr << "\n(marked as " << printType(oldType) << ", should be " << printType(newType) << ")\n";
           info.fail(ss.str(), curr, getFunction());
         }
         curr->type = oldType;
       }
+      // check if a node is a duplicate - expressions must not be seen more than once
+      bool inserted;
+      std::tie(std::ignore, inserted) = seen.insert(curr);
+      if (!inserted) {
+        std::ostringstream ss;
+        ss << "expression seen more than once in the tree in " << scope << " on " << curr << '\n';
+        info.fail(ss.str(), curr, getFunction());
+      }
     }
   };
   BinaryenIRValidator binaryenIRValidator(info);
@@ -920,23 +913,15 @@ static void validateBinaryenIR(Module& wasm, ValidationInfo& info) {
 // Main validator class
 
 static void validateImports(Module& module, ValidationInfo& info) {
-  for (auto& curr : module.imports) {
-    if (curr->kind == ExternalKind::Function) {
-      if (info.validateWeb) {
-        auto* functionType = module.getFunctionType(curr->functionType);
-        info.shouldBeUnequal(functionType->result, i64, curr->name, "Imported function must not have i64 return type");
-        for (Type param : functionType->params) {
-          info.shouldBeUnequal(param, i64, curr->name, "Imported function must not have i64 parameters");
-        }
+  ModuleUtils::iterImportedFunctions(module, [&](Function* curr) {
+    if (info.validateWeb) {
+      auto* functionType = module.getFunctionType(curr->type);
+      info.shouldBeUnequal(functionType->result, i64, curr->name, "Imported function must not have i64 return type");
+      for (Type param : functionType->params) {
+        info.shouldBeUnequal(param, i64, curr->name, "Imported function must not have i64 parameters");
       }
     }
-    if (curr->kind == ExternalKind::Table) {
-      info.shouldBeTrue(module.table.imported, curr->name, "Table import record exists but table is not marked as imported");
-    }
-    if (curr->kind == ExternalKind::Memory) {
-      info.shouldBeTrue(module.memory.imported, curr->name, "Memory import record exists but memory is not marked as imported");
-    }
-  }
+  });
 }
 
 static void validateExports(Module& module, ValidationInfo& info) {
@@ -951,17 +936,13 @@ static void validateExports(Module& module, ValidationInfo& info) {
       }
     }
   }
-  std::set<Name> exportNames;
+  std::unordered_set<Name> exportNames;
   for (auto& exp : module.exports) {
     Name name = exp->value;
     if (exp->kind == ExternalKind::Function) {
-      Import* imp;
-      info.shouldBeTrue(module.getFunctionOrNull(name) ||
-                        ((imp = module.getImportOrNull(name)) && imp->kind == ExternalKind::Function), name, "module function exports must be found");
+      info.shouldBeTrue(module.getFunctionOrNull(name), name, "module function exports must be found");
     } else if (exp->kind == ExternalKind::Global) {
-      Import* imp;
-      info.shouldBeTrue(module.getGlobalOrNull(name) ||
-                        ((imp = module.getImportOrNull(name)) && imp->kind == ExternalKind::Global), name, "module global exports must be found");
+      info.shouldBeTrue(module.getGlobalOrNull(name), name, "module global exports must be found");
     } else if (exp->kind == ExternalKind::Table) {
       info.shouldBeTrue(name == Name("0") || name == module.table.name, name, "module table exports must be found");
     } else if (exp->kind == ExternalKind::Memory) {
@@ -976,19 +957,19 @@ static void validateExports(Module& module, ValidationInfo& info) {
 }
 
 static void validateGlobals(Module& module, ValidationInfo& info) {
-  for (auto& curr : module.globals) {
+  ModuleUtils::iterDefinedGlobals(module, [&](Global* curr) {
     info.shouldBeTrue(curr->init != nullptr, curr->name, "global init must be non-null");
     info.shouldBeTrue(curr->init->is<Const>() || curr->init->is<GetGlobal>(), curr->name, "global init must be valid");
     if (!info.shouldBeEqual(curr->type, curr->init->type, curr->init, "global init must have correct type") && !info.quiet) {
       info.getStream(nullptr) << "(on global " << curr->name << ")\n";
     }
-  }
+  });
 }
 
 static void validateMemory(Module& module, ValidationInfo& info) {
   auto& curr = module.memory;
   info.shouldBeFalse(curr.initial > curr.max, "memory", "memory max >= initial");
-  info.shouldBeTrue(curr.max <= Memory::kMaxSize, "memory", "max memory must be <= 4GB");
+  info.shouldBeTrue(!curr.hasMax() || curr.max <= Memory::kMaxSize, "memory", "max memory must be <= 4GB, or unlimited");
   info.shouldBeTrue(!curr.shared || curr.hasMax(), "memory", "shared memory must have max size");
   if (curr.shared) info.shouldBeTrue(info.features & Feature::Atomics, "memory", "memory is shared, but atomics are disabled");
   for (auto& segment : curr.segments) {
@@ -1010,7 +991,7 @@ static void validateTable(Module& module, ValidationInfo& info) {
     info.shouldBeEqual(segment.offset->type, i32, segment.offset, "segment offset should be i32");
     info.shouldBeTrue(checkOffset(segment.offset, segment.data.size(), module.table.initial * Table::kPageSize), segment.offset, "segment offset should be reasonable");
     for (auto name : segment.data) {
-      info.shouldBeTrue(module.getFunctionOrNull(name) || module.getImportOrNull(name), name, "segment name should be valid");
+      info.shouldBeTrue(module.getFunctionOrNull(name), name, "segment name should be valid");
     }
   }
 }
